# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


# Future Modules:
from __future__ import annotations

# Built-in Modules:
import ctypes
import os
from typing import Optional, Union

# Third-party Modules:
import win32com.client  # type: ignore[import]
from pywintypes import com_error as ComError  # type: ignore[import]

# Local Modules:
from . import LIB_DIRECTORY, SYSTEM_ARCHITECTURE
from .base import BaseSpeech


# SAPI constants
SPF_ASYNC: int = 1  # Specifies that the Speak call should be asynchronous.
SPF_PURGEBEFORESPEAK: int = 2  # Purges all pending speak requests prior to this speak call.
SPF_IS_NOT_XML: int = 16  # The input text will not be parsed for XML markup.


class Speech(BaseSpeech):
	def __init__(self) -> None:
		self.find_window: ctypes._NamedFuncPointer = ctypes.WinDLL("user32").FindWindowW
		self.find_window.argtypes = [ctypes.c_wchar_p, ctypes.c_wchar_p]
		self.find_window.restype = ctypes.c_void_p
		self.nvda: ctypes.WinDLL
		self.sa: ctypes.WinDLL
		if SYSTEM_ARCHITECTURE == "32bit":
			self.nvda = ctypes.windll.LoadLibrary(os.path.join(LIB_DIRECTORY, "nvdaControllerClient32.dll"))
			self.sa = ctypes.windll.LoadLibrary(os.path.join(LIB_DIRECTORY, "SAAPI32.dll"))
		else:  # pragma: no cover
			self.nvda = ctypes.windll.LoadLibrary(os.path.join(LIB_DIRECTORY, "nvdaControllerClient64.dll"))
			self.sa = ctypes.windll.LoadLibrary(os.path.join(LIB_DIRECTORY, "SAAPI64.dll"))
		self.nvda.nvdaController_brailleMessage.argtypes = (ctypes.c_wchar_p,)
		self.nvda.nvdaController_speakText.argtypes = (ctypes.c_wchar_p,)
		self.sa.SA_BrlShowTextW.argtypes = (ctypes.c_wchar_p,)
		self.sa.SA_SayW.argtypes = (ctypes.c_wchar_p,)

	@property
	def sapi(  # type: ignore[misc, no-any-unimported]
		self,
	) -> Union[win32com.client.CDispatch, None]:  # pragma: no cover
		"""The SAPI COM object."""
		try:
			return win32com.client.Dispatch("SAPI.SpVoice")
		except ComError:
			return None

	@property
	def jfw(  # type: ignore[misc, no-any-unimported]
		self,
	) -> Union[win32com.client.CDispatch, None]:  # pragma: no cover
		"""The JFW COM object."""
		try:
			return win32com.client.Dispatch("FreedomSci.JawsApi")
		except ComError:
			return None

	def jfw_braille(self, text: str) -> None:
		"""
		Brailles text using JFW.

		Args:
			text: The text to braille.
		"""
		self.jfw_output(text, braille=True)

	def jfw_output(
		self,
		text: str,
		braille: Optional[bool] = None,
		speak: Optional[bool] = None,
		interrupt: Optional[bool] = None,
	) -> None:
		"""
		Outputs text using JFW.

		Args:
			text: The output text.
			braille: Output text in braille.
			speak: Output text using speech.
			interrupt: True if the speech should be silenced before speaking.
		"""
		jfw: Union[win32com.client.CDispatch, None] = self.jfw  # type: ignore[no-any-unimported]
		if jfw is not None:
			if speak:
				jfw.SayString(text, int(bool(interrupt)))
			if braille:
				jfw.RunFunction('BrailleString("{text}")'.format(text=text.replace('"', "'")))

	def jfw_running(self) -> bool:
		"""
		Determines if JFW is running.

		Returns:
			True if JFW is running, False otherwise.
		"""
		return bool(self.find_window("JFWUI2", None))

	def jfw_say(self, text: str, interrupt: Optional[bool] = None) -> None:
		"""
		Speak text using JFW.

		Args:
			text: The text to be spoken.
			interrupt: True if the speech should be silenced before speaking.
		"""
		self.jfw_output(text, speak=True, interrupt=interrupt)

	def jfw_silence(self) -> None:
		"""Cancels JFW speech and flushes the speech buffer."""
		jfw: Union[win32com.client.CDispatch, None] = self.jfw  # type: ignore[no-any-unimported]
		if jfw is not None:
			jfw.StopSpeech()

	def nvda_braille(self, text: str) -> None:
		"""
		Brailles text using NVDA.

		Args:
			text: The text to braille.
		"""
		self.nvda.nvdaController_brailleMessage(text)

	def nvda_output(self, text: str, interrupt: Optional[bool] = None) -> None:
		"""
		Outputs text using NVDA.

		Args:
			text: The output text.
			interrupt: True if the speech should be silenced before speaking.
		"""
		self.nvda_say(text, interrupt)
		self.nvda_braille(text)

	def nvda_running(self) -> bool:
		"""
		Determines if NVDA is running.

		Returns:
			True if NVDA is running, False otherwise.
		"""
		return bool(self.nvda.nvdaController_testIfRunning() == 0)

	def nvda_say(self, text: str, interrupt: Optional[bool] = None) -> None:
		"""
		Speak text using NVDA.

		Args:
			text: The text to be spoken.
			interrupt: True if the speech should be silenced before speaking.
		"""
		if interrupt:
			self.nvda_silence()
		self.nvda.nvdaController_speakText(text)

	def nvda_silence(self) -> None:
		"""Cancels NVDA speech and flushes the speech buffer."""
		self.nvda.nvdaController_cancelSpeech()

	def sa_braille(self, text: str) -> None:
		"""
		Brailles text using System Access.

		Args:
			text: The text to braille.
		"""
		self.sa.SA_BrlShowTextW(text)

	def sa_output(self, text: str, interrupt: Optional[bool] = None) -> None:
		"""
		Outputs text using System Access.

		Args:
			text: The output text.
			interrupt: True if the speech should be silenced before speaking.
		"""
		self.sa_say(text, interrupt)
		self.sa_braille(text)

	def sa_running(self) -> bool:
		"""
		Determines if System Access is running.

		Returns:
			True if System Access is running, False otherwise.
		"""
		return bool(self.sa.SA_IsRunning())

	def sa_say(self, text: str, interrupt: Optional[bool] = None) -> None:
		"""
		Speak text using System Access.

		Args:
			text: The text to be spoken.
			interrupt: True if the speech should be silenced before speaking.
		"""
		if interrupt:
			self.sa_silence()
		self.sa.SA_SayW(text)

	def sa_silence(self) -> None:
		"""Cancels System Access speech and flushes the speech buffer."""
		self.sa.SA_StopAudio()

	def sapi_say(self, text: str, interrupt: Optional[bool] = None) -> None:
		"""
		Speak text using SAPI.

		Args:
			text: The text to be spoken.
			interrupt: True if the speech should be silenced before speaking.
		"""
		if self.sapi is not None:
			if interrupt:
				self.sapi.Speak(text, SPF_ASYNC | SPF_PURGEBEFORESPEAK | SPF_IS_NOT_XML)
			else:
				self.sapi.Speak(text, SPF_ASYNC | SPF_IS_NOT_XML)

	def sapi_silence(self) -> None:
		"""Cancels SAPI speech and flushes the speech buffer."""
		if self.sapi is not None:
			self.sapi.Speak("", SPF_ASYNC | SPF_PURGEBEFORESPEAK | SPF_IS_NOT_XML)

	def braille(self, text: str) -> None:
		if self.nvda_running():
			self.nvda_braille(text)
		elif self.sa_running():
			self.sa_braille(text)
		elif self.jfw_running():
			self.jfw_braille(text)

	def output(self, text: str, interrupt: Optional[bool] = None) -> None:
		if self.nvda_running():
			self.nvda_output(text, interrupt)
		elif self.sa_running():
			self.sa_output(text, interrupt)
		elif self.jfw_running():
			self.jfw_output(text, braille=True, speak=True, interrupt=interrupt)
		else:
			self.sapi_say(text, interrupt)

	def say(self, text: str, interrupt: Optional[bool] = None) -> None:
		if self.nvda_running():
			self.nvda_say(text, interrupt)
		elif self.sa_running():
			self.sa_say(text, interrupt)
		elif self.jfw_running():
			self.jfw_say(text, interrupt)
		else:
			self.sapi_say(text, interrupt)

	def silence(self) -> None:
		if self.nvda_running():
			self.nvda_silence()
		elif self.sa_running():
			self.sa_silence()
		elif self.jfw_running():
			self.jfw_silence()
		else:
			self.sapi_silence()

	def speaking(self) -> bool:
		if self.nvda_running() or self.sa_running() or self.jfw_running():
			# None of the screen reader APIs support retrieving speaking status.
			return False
		elif self.sapi is not None:
			return bool(self.sapi.Status.RunningState != 1)
		return False



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


# Future Modules:
from __future__ import annotations

# Built-in Modules:
from typing import Optional

# Third-party Modules:
from Cocoa import NSSpeechSynthesizer  # type: ignore[import]

# Local Modules:
from .base import BaseSpeech


class Speech(BaseSpeech):
	def __init__(self) -> None:
		# Allocate and initialize the default TTS.
		self.darwin = NSSpeechSynthesizer.alloc().init()

	def braille(self, text: str) -> None:
		pass

	def output(self, text: str, interrupt: Optional[bool] = None) -> None:
		self.say(text, interrupt)
		self.braille(text)

	def say(self, text: str, interrupt: Optional[bool] = None) -> None:
		if interrupt:
			self.silence()
		self.darwin.startSpeakingString_(text)

	def silence(self) -> None:
		self.darwin.stopSpeaking()

	def speaking(self) -> bool:
		return bool(self.darwin.isSpeaking())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


# Future Modules:
from __future__ import annotations

# Built-in Modules:
import _imp
import os
import sys
from typing import Union


def get_freezer() -> Union[str, None]:
	"""
	Determines the name of the library used to freeze the code.

	Note:
		https://github.com/blackmagicgirl/ktools/blob/master/ktools/utils.py

	Returns:
		The name of the library or None.
	"""
	frozen: Union[str, bool, None] = getattr(sys, "frozen", None)
	if frozen and hasattr(sys, "_MEIPASS"):
		return "pyinstaller"
	elif frozen is True:
		return "cx_freeze"
	elif frozen in ("windows_exe", "console_exe", "dll"):
		return "py2exe"
	elif frozen == "macosx_app":
		return "py2app"
	elif hasattr(sys, "importers"):
		return "old_py2exe"
	elif _imp.is_frozen("__main__"):
		return "tools/freeze"
	elif isinstance(frozen, str):
		return f"unknown {frozen}"
	return None


def is_frozen() -> bool:
	"""
	Determines whether the program is running from a frozen copy or from source.

	Returns:
		True if frozen, False otherwise.
	"""
	return bool(get_freezer())


def get_directory_path(*args: str) -> str:
	"""
	Retrieves the path of the directory where the program is located.

	Args:
		*args: Positional arguments to be passed to os.join after the directory path.

	Returns:
		The path.
	"""
	if is_frozen():
		path = os.path.dirname(sys.executable)
	else:
		path = os.path.join(os.path.dirname(__file__))
	return os.path.realpath(os.path.join(path, *args))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


# Future Modules:
from __future__ import annotations

# Built-in Modules:
import pathlib
from typing import List

# Third-party Modules:
from setuptools import setup  # type: ignore[import]


NAME: str = "Speechlight"
DESCRIPTION: str = (
	"A lightweight Python library providing a common interface to multiple TTS and screen reader APIs."
)
KEYWORDS: str = "blind jaws jfw nvda speech tts screenreader screen reader"
AUTHOR: str = "Nick Stockton"
AUTHOR_EMAIL: str = "nstockton@users.noreply.github.com"
VERSION: str = "1.5"
URL: str = "https://github.com/nstockton/speechlight"
# The directory containing this file
HERE: pathlib.Path = pathlib.Path(__file__).parent
README: str = (HERE / "README.md").read_text()
REQUIREMENTS: List[str] = [
	"pywin32 ; platform_system=='Windows'",
]


setup(
	python_requires=">=3.7",
	name=NAME,
	keywords=KEYWORDS,
	author=AUTHOR,
	author_email=AUTHOR_EMAIL,
	version=VERSION,
	description=DESCRIPTION,
	long_description=README,
	long_description_content_type="text/markdown",
	url=URL,
	package_dir={"speechlight": "speechlight"},
	packages=["speechlight"],
	package_data={"speechlight": ["py.typed", "speech_libs/*"]},
	zip_safe=False,
	setup_requires=REQUIREMENTS,
	install_requires=REQUIREMENTS,
	scripts=[],
	license="Mozilla Public License 2.0 (MPL 2.0)",
	platforms="Posix; MacOS X; Windows",
	classifiers=[
		"License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
		"Programming Language :: Python :: 3 :: Only",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
		"Programming Language :: Python",
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: Developers",
		"Operating System :: OS Independent",
		"Topic :: Adaptive Technologies",
		"Topic :: Software Development :: Libraries",
	],
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


# Future Modules:
from __future__ import annotations

# Built-in Modules:
from abc import ABC, abstractmethod
from typing import Optional


class BaseSpeech(ABC):
	@abstractmethod
	def braille(self, text: str) -> None:
		"""
		Brailles text.

		Args:
			text: The text to be brailled.
		"""

	@abstractmethod
	def output(self, text: str, interrupt: Optional[bool] = None) -> None:
		"""
		Speaks and brailles text.

		Args:
			text: The output text.
			interrupt: True if the speech should be silenced before speaking.
		"""

	@abstractmethod
	def say(self, text: str, interrupt: Optional[bool] = None) -> None:
		"""
		Speaks text.

		Args:
			text: The text to be spoken.
			interrupt: True if the speech should be silenced before speaking.
		"""

	@abstractmethod
	def silence(self) -> None:
		"""Cancels speech and flushes the speech buffer."""

	@abstractmethod
	def speaking(self) -> bool:
		"""
		Determines if text is currently being spoken.

		Returns:
			True if text is currently being spoken, False otherwise.
		"""



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


# Future Modules:
from __future__ import annotations

# Built-in Modules:
import platform
from typing import TYPE_CHECKING, List

# Local Modules:
from .utils import get_directory_path


SYSTEM_ARCHITECTURE: str = platform.architecture()[0]
SYSTEM_PLATFORM: str = platform.system()
LIB_DIRECTORY: str = get_directory_path("speech_libs")


if TYPE_CHECKING:  # pragma: no cover
	from .dummy import Speech
else:  # pragma: no cover
	if SYSTEM_PLATFORM == "Windows":
		from .windows import Speech
	elif SYSTEM_PLATFORM == "Darwin":
		from .darwin import Speech
	else:
		from .dummy import Speech

speech: Speech = Speech()
__all__: List[str] = ["Speech", "speech"]
